#ifndef BASEPAIRS_HH
#define BASEPAIRS_HH

#include <vector>
#include <ext/hash_map>
#include <assert.h>

#include "base.hh"
#include "sequence.hh"

#define SPACE_OPT /* save some space */

/*!
   parameters for the basepairs class

   contains the necessary information for determining the
   significant basepairs and computing their weights
*/
class BasePairsParams {
protected:
	const int structure_weight; //!< weight for balancing structure contribution to sequence contribution in the score
  
  double prob_exp_f(int seqlen) const {return 1.0/(2.0*seqlen);} //!<expected probability of a base pair (null-model)
  
  const double min_prob; //!< only pairs with a probability of at least min_prob are taken into account
  const double exp_prob; //!< expected probability
  const bool opt_exp_prob;
public:
  BasePairsParams(int _structure_weight, double _min_prob, double _exp_prob, bool _opt_exp_prob):
    structure_weight(_structure_weight),
    min_prob(_min_prob), exp_prob(_exp_prob),
    opt_exp_prob(_opt_exp_prob)
  {}
};

//! represents a base pair; stores base pair index and weight
class Arc {
public:
    int idx;
    score_t w;

    Arc(int _idx, score_t _w): idx(_idx), w(_w) {}
};

//! represents a base pair in a left adjacency list;
//! stores base pair index, position of right end, and weight
class LeftAdjEntry {
public:
    int idx;
    int right; 
    score_t w;
    LeftAdjEntry(int _idx, int _right, score_t _w): idx(_idx), right(_right), w(_w) {}
};

//! symmetric class to LeftAdjEntry
class RightAdjEntry {
public:
    int idx;
    int left; 
    score_t w;
    RightAdjEntry(int _idx, int _left, score_t _w): idx(_idx), left(_left), w(_w) {}
};

#ifdef SPACE_OPT

/* in order to save some space we use a hash_map for storing the entries of the sparse
   matrices that allow fast acces to the basepairs (i,j) in a structure.
*/

namespace __gnu_cxx {
// define a new hash function for pairs of ints
template<>
    struct hash<std::pair<size_t,size_t> >
    {
      size_t
      operator()(std::pair<size_t,size_t> p) const
			{ return p.first<<(sizeof(size_t)/2) | p.second; }
    };
}

//! represents a sparse 2D matrix of entries val_t with default value deflt
//! implements the matrix by a hash map
template <class val_t,int deflt>
class SparseMatrix {
private:
	typedef std::pair<size_t,size_t> key_t;
	typedef __gnu_cxx::hash_map<key_t,val_t > map_t;
	
	const val_t def;
	
	map_t the_map;
public:
	SparseMatrix() : the_map(),def(deflt) {}

	const val_t & operator() (size_t i, size_t j) const {
		 typename map_t::const_iterator it = the_map.find(key_t(i,j));
		 if ( it == the_map.end() ) 
			 return def;
		 else 
			 return it->second;
	}
	void set(size_t i, size_t j, const val_t &val) {
		the_map[std::pair<int,int>(i,j)]=val;
	}
};
	
#endif

// ============================================================
// class BasePairs
/*! stores and maintains the list of base pairs together with their score
  contributions in arc matches.
*/
// ============================================================
class BasePairs:
    BasePairsParams 
{
public:
	
	/* types for data structures for the access of an arc in the structure,
	   by its right end, its left end, or left and right end 
	
	   when using SPACE_OPT,
	   the latter access structure is implemented as a hash map,
	   otherwise, as array with O(n^2) space
	*/
	
    typedef std::vector<Arc> arc_vec_t;
    typedef std::vector<LeftAdjEntry> LeftAdjList;
    typedef std::vector<RightAdjEntry> RightAdjList;



#ifdef SPACE_OPT
    typedef SparseMatrix<int,-1> arc_matrix_t;
	typedef SparseMatrix<double,0> arc_prob_matrix_t;
#else
    typedef std::vector<std::vector<int> > arc_matrix_t;
    typedef std::vector<std::vector<double> > arc_prob_matrix_t;
#endif


private:
    std::vector<LeftAdjList> left_;
    std::vector<RightAdjList> right_;
    arc_vec_t arc_vec_;

    arc_matrix_t arcs_;
    arc_prob_matrix_t arc_probs_; //!< array for all arc probabilities (used for merging in pp-output)
    
    void resize(uint s);
    
    void sortAdjLists();
    
    int s_; //!< the sequence
    int len_; //!< the sequence length
public:
    BasePairs(const BasePairsParams &bpp):BasePairsParams(bpp), s_(0), len_(0) {}
    //! adds a basepair (i,j) with weight w
	//! also maintains the basepair access data structures
    void insert(int i, int j, score_t w);
  
	//! returns probability of basepair (i,j)
    double get_arc_prob(int i, int j) const {
      assert(i<j); 
#ifdef SPACE_OPT
	  return arc_probs_(i,j);
#else
	  return arc_probs_[i][j];
#endif
	}

	//! returns the list of arcs with right end i
    const LeftAdjList & left_adjlist(int i) const {return left_[i];}
	
    //! returns the list of arcs with left end i
    const RightAdjList & right_adjlist(int i) const {return right_[i];}

#ifdef SPACE_OPT
    //! accesses basepair by (i,j)
    const Arc &arc(int i,int j) const {return arc_vec_[arcs_(i,j)];}
    //! returns whether basepair (i,j) exitsts
	bool exists_arc(int i,int j) const {return -1 != arcs_(i,j);}
#else
    const Arc &arc(int i,int j) const {return arc_vec_[arcs_[i][j]];}
    bool exists_arc(int i,int j) const {return -1 != arcs_[i][j];}
#endif

	//! returns number of basepairs in the object
    int num_bps() const {return arc_vec_.size();}
	
	//! returns weight that corresponds to probability p
    int probToWeight(double p) const;
	
    //! read baepairs and sequence from a pp-format file
    //! pp is a proprietary format of LocARNA
    //! which starts with the sequence and then simply
	//! lists the arcs (i,j) with their probabilities p
	void readPP(char *filename, sequence_t &seq);
    
	//! read baepairs and sequence from a dp-format ps file
    //! dp is written by RNAfold -p
	void readPS(char *filename, sequence_t &seq);
    
	//! read baepairs and sequence from a file
	//! try to autodetect file format
	void read(char *filename, sequence_t &seq);

	double prob_exp() const; //!< return expected probability
    double prob_min() const; //!< return minimal probability

    //! returns the base pair parameters that were used to init the object
	const BasePairsParams &get_params() const {return *this;};
};

typedef BasePairs Bps; 

#endif
